package server.network;

import java.security.MessageDigest;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.util.Arrays;

import javax.crypto.SecretKey;

import common.color.TextColor;
import common.init.Config;
import common.log.Log;
import common.net.util.concurrent.Future;
import common.net.util.concurrent.GenericFutureListener;
import common.network.ILoginHandler;
import common.network.IPlayer;
import common.network.NetConnection;
import common.packet.LPacketChallenge;
import common.packet.LPacketPassword;
import common.packet.LPacketPubkey;
import common.packet.LPacketResponse;
import common.packet.LPacketStartEncrypt;
import common.packet.RPacketChallenge;
import common.packet.RPacketDisconnect;
import common.packet.RPacketRequestEncrypt;
import common.packet.RPacketResponse;
import common.packet.RPacketServerConfig;
import common.util.EncryptUtil;
import server.Server;

public class LoginHandler implements ILoginHandler
{
    private static enum LoginState {
        INIT, ENCRYPT, PROOF, PASSWORD, CHALLENGE, AUTHENTICATED, ACCEPTED;
    }
    
    private static final SecureRandom TOKEN_RNG = new SecureRandom();
    
    private final Server server;
    public final NetConnection netManager;
    
    private LoginState state = LoginState.INIT;
    private int timer;
    private String loginUser;
    private String loginPass;
    private byte[] loginToken;
    private PublicKey loginKey;

    public LoginHandler(Server server, NetConnection netManager)
    {
        this.netManager = netManager;
        this.server = server;
    }

    public void closeConnection(String reason)
    {
        try
        {
            Log.NETWORK.info("Trenne " + this.getConnectionInfo());
            this.netManager.sendPacket(new RPacketDisconnect(reason));
            this.netManager.closeChannel(reason);
        }
        catch (Exception exception)
        {
            Log.NETWORK.error(exception, "Fehler beim Trennen des Spielers beim Login");
        }
    }
    
    public void onDisconnect(String reason)
    {
        Log.NETWORK.info(this.getConnectionInfo() + " wurde beim Login getrennt: " + TextColor.stripCodes(reason));
    }

    public String getConnectionInfo()
    {
        return this.loginUser != null ? (this.loginUser + " (" + this.netManager.getCutAddress()
        		+ ")") : this.netManager.getCutAddress();
    }
    
    public void update()
    {
        if(this.state == LoginState.AUTHENTICATED)
        	this.tryAdvance();
        if(this.timer++ == 600)
            this.closeConnection("Anmeldung dauerte zu lange");
    }

    private void tryAdvance() {
        String kick = this.checkPlayer();
        if(kick != null) {
        	this.closeConnection(kick);
        	return;
        }
        this.server.addPlayer(this.netManager, this.loginUser);
        this.state = LoginState.ACCEPTED;
    }
    
    private String checkPlayer() {
		if(this.server.getPlayer(this.loginUser) != null)
			return "Nutzername '" + this.loginUser + "' ist bereits vergeben";
		User user = this.server.getUser(this.loginUser);
		if(Config.authenticate) {
			if(user == null) {
				if(!Config.register)
					return "Anmeldung neuer Accounts ist auf diesem Server deaktiviert (Whitelisted)";
				if(Config.playerLimit > 0 && this.server.getPlayers().size() >= Config.playerLimit)
					return String.format("Der Server ist voll (%d/%d)!", this.server.getPlayers().size(), Config.playerLimit);
				if(this.loginKey != null) {
					this.server.addUser(user = new User(this.loginUser));
					user.setPubkey(this.loginKey);
					Log.NETWORK.info(this.loginUser + " registrierte sich mit Pubkey");
				}
				else {
					if(this.loginPass == null || this.loginPass.length() == 0)
						return "Ein neues Passwort ist erforderlich um diesen Server zu betreten (mindestens " + Config.minPassLength + " Zeichen)";
					if(this.loginPass.length() < Config.minPassLength)
						return "Passwort ist zu kurz, mindestens " + Config.minPassLength + " Zeichen";
					this.server.addUser(user = new User(this.loginUser));
					user.setPasswordHash(EncryptUtil.hashPassword(this.loginPass));
					Log.NETWORK.info(this.loginUser + " registrierte sich mit Passwort");
				}
			}
			else if((user.getPasswordHash() == null && user.getPubkey() == null) || (user.getPubkey() != null ? !user.getPubkey().equals(this.loginKey) : 
					(this.loginPass == null || !MessageDigest.isEqual(EncryptUtil.hashPassword(this.loginPass, user.getPasswordHash().second()), user.getPasswordHash().first())))) {
				return this.loginKey != null ? "Falscher Pubkey" : "Falsches Passwort";
			}
			else {
				Log.NETWORK.info(this.loginUser + " loggte sich mit " + (this.loginKey != null ? "Pubkey" : "Passwort") + " ein");
			}
		}
		if(Config.playerLimit > 0 && this.server.getPlayers().size() >= Config.playerLimit && (user == null || !user.isAdmin()))
			return String.format("Der Server ist voll (%d/%d)!", this.server.getPlayers().size(), Config.playerLimit);
		return null;
    }

    public void sendLoginPacket() {
		if(this.state != LoginState.INIT)
			throw new IllegalStateException("Unerwartetes Handshake-Paket");
		if(Config.encrypt) {
			this.state = LoginState.ENCRYPT;
			this.loginToken = new byte[4];
			TOKEN_RNG.nextBytes(this.loginToken);
	    	this.netManager.sendPacket(new RPacketRequestEncrypt(this.server.getPublicKey(), this.loginToken));
		}
		else {
			this.state = LoginState.PASSWORD;
			this.netManager.sendPacket(new RPacketServerConfig(Config.accessRequired, Config.authenticate, Config.authenticate && Config.passwordAuth, false));
		}
    }
    
	public void processEncryption(LPacketStartEncrypt packet) {
		if(this.state != LoginState.ENCRYPT)
			throw new IllegalStateException("Unerwartetes Verschlüsselungs-Paket");
        PrivateKey pkey = this.server.getPrivateKey();
        if(!Arrays.equals(this.loginToken, packet.getToken(pkey)))
            throw new IllegalStateException("Fehlerhaftes Token");
        SecretKey key = packet.getKey(pkey);
        this.netManager.startEncryption(key);
		this.state = LoginState.PROOF;
	}

	public void processChallenge(LPacketChallenge packet) {
		if(this.state != LoginState.PROOF)
			throw new IllegalStateException("Unerwartetes Anforderungs-Paket");
		this.state = LoginState.PASSWORD;
		this.netManager.sendPacket(new RPacketResponse(packet.getToken(this.server.getPrivateKey())), new GenericFutureListener < Future <? super Void >> () {
            public void operationComplete(Future <? super Void > u) throws Exception {
                LoginHandler.this.netManager.sendPacket(new RPacketServerConfig(Config.accessRequired, Config.authenticate, Config.authenticate && Config.passwordAuth, 
	    			Config.authenticate && Config.pubkeyAuth));
            }
        });
	}
	
	private boolean checkAccess(String access) {
		if(Config.accessRequired) {
	    	if(Config.password.length() < 8) {
	    		this.closeConnection("Es ist kein Zugangspasswort für diesen Server konfiguriert");
	    		return false;
			}
	    	if(!Config.password.equals(access)) {
	    		this.closeConnection("Falsches Zugangspasswort");
	    		return false;
			}
		}
		return true;
	}
    
    public void processPassword(LPacketPassword packet) {
		if(this.state != LoginState.PASSWORD)
			throw new IllegalStateException("Unerwartetes Passwort-Paket");
		this.loginUser = packet.getUser();
        if(this.loginUser.isEmpty() || !IPlayer.isValidUser(this.loginUser))
        	throw new IllegalStateException("Ungültiger Nutzername!");
		if(!Config.passwordAuth && Config.authenticate) {
			this.closeConnection("Dieser Server " + (Config.pubkeyAuth && Config.encrypt ? "benötigt einen öffentlichen Schlüssel zur Authentifizierung" : "hat keine Authentifizierungsmethode konfiguriert"));
			return;
		}
		if(!this.checkAccess(packet.getAccess()))
			return;
		if(Config.authenticate)
			this.loginPass = packet.getPassword();
		this.state = LoginState.AUTHENTICATED;
	}

	public void processPubkey(LPacketPubkey packet) {
		if(this.state != LoginState.PASSWORD)
			throw new IllegalStateException("Unerwartetes Pubkey-Paket");
		this.loginUser = packet.getUser();
        if(this.loginUser.isEmpty() || !IPlayer.isValidUser(this.loginUser))
        	throw new IllegalStateException("Ungültiger Nutzername!");
		if((!Config.pubkeyAuth || !Config.encrypt) && Config.authenticate) {
			this.closeConnection("Dieser Server " + (Config.passwordAuth ? "benötigt ein Passwort zur Authentifizierung" : "hat keine Authentifizierungsmethode konfiguriert"));
			return;
		}
		if(!this.checkAccess(packet.getAccess()))
			return;
    	if(Config.authenticate) {
			this.loginKey = packet.getKey();
			this.loginToken = new byte[32];
			TOKEN_RNG.nextBytes(this.loginToken);
			this.netManager.sendPacket(new RPacketChallenge(this.loginKey, this.loginToken));
			this.state = LoginState.CHALLENGE;
    	}
    	else {
    		this.state = LoginState.AUTHENTICATED;
    	}
	}

	public void processResponse(LPacketResponse packet) {
		if(this.state != LoginState.CHALLENGE)
			throw new IllegalStateException("Unerwartetes Beweis-Paket");
		if(!Arrays.equals(this.loginToken, packet.getToken()))
            throw new IllegalStateException("Fehlerhaftes Beweis-Token");
		this.state = LoginState.AUTHENTICATED;
	}
}
